/*
 *	Name: Hangman game.
 *
 *	Version: 1.0
 *
 *	Desc:
 *		This is a playable copy of hangman game.
 *		Its sole purpose is to demonstrate basic curses package usage.
 *		This program should compile on any system that has Curses.
 *
 *	Author: Antonio Bonifati	(antonio.b@infinito.it).
 *
 *	Date: 26.Nov.2000
 *
 */

#include <curses.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <signal.h>

/* -----------------------Configure here----------------------- */
/* change here the program version */
#define VERSION		"hangman-1.0"
/* name of the user data file - this is searched first */
#define HOMEDATAFILENAME        ".hangman.words"
/* name of the default data file */
#define DATAFILENAME    "/usr/share/games/hangman.words"
/* max length of line of DATAFILENAME */
#define MAXLINELEN	200
/* field delimiter in file DATAFILENAME */
#define DELIMITER	"|"
/* char used to hidden a not yet guessed char */
#define HIDDENCHAR	'_'
/* the total number of tries will be word length + ERRORS.
   this is the default value which can also be changed
   via the command line option -t */
#define ERRORS	1
/* ------------------------------------------------------------ */

/* game variables and settings */
char *word, *suggestion, triedchars[MAXLINELEN], *datafile, *lastmsg;
int tries, triesleft, chosenlen, errors=ERRORS;
bool quite, testword;
/*******************************/

/* dies with the name of the program and
   an error message. return the error
   number */
void die(char *what)
{
   extern char* prog;
   endwin();
   fprintf(stderr, "%s: %s\n", prog, what);
   perror(prog);
   exit(errno);
}

/* hack to return the file size;
   don't changes the position
   indicator */
long filesize(FILE *f)
{
   long size, curpos=ftell(f);
   fseek(f, 0, SEEK_END);
   size=ftell(f);
   fseek(f, curpos, SEEK_SET);
   return size;
}

/* removes a newline at the end of a string
   cannot be used for constant strings */
char *chomp(char *s)
{
   if (s && *s)
   {
      size_t last=strlen(s)-1;
      if (s[last]=='\n')
         s[last]='\0';
   }
   return s;
}

/* similar to strchr(3),
   but compares are case-insensitive */
const char *strcasechr(const char *s, int c)
{
   if (s)
      for (c=tolower(c); (tolower(*s) != c); s++)
	if (!*s) return NULL;
   return s;
}

char *my_strtok(char *s, const char *delim)
{
   static char *p;

   for (!s?(s=p):(p=s); !strchr(delim, *p); p++);
   if (*p) *p++='\0';

   return s;
}

/* similar to strlen but includes only
   letter characters ([A-Za-z]) in the count */
size_t stralphalen(const char *s)
{
   size_t alphalen;

   for (alphalen=0; *s; s++)
      if (isalpha(*s))
	 alphalen++;
   return alphalen;
}

void getrandline(char *l, int n)
{
   long size, randpos;
   FILE *f;
   char fname[200];

   /* try to open [FILE] (if specified) */
   /* if this fails, die with error */
   if (datafile)
   {
      f=fopen(datafile, "r");
      if (f == NULL) die(datafile);
   }
   else
   {
      /* try to open $HOME/HOMEDATAFILENAME */
      strcpy(fname, (char *)getenv("HOME"));
      strcat(fname, "/");
      strcat(fname, HOMEDATAFILENAME);
      f=fopen(fname, "r");

      /* if this fails, try to open DATAFILENAME */
      if (f == NULL)
      {
	 f=fopen(DATAFILENAME, "r");
	 if (f == NULL) die("trying to open " DATAFILENAME);
      }
   }

   size=filesize(f);

   do
   {
      /* get a random value in the range [0..size-1] */
      randpos=rand()%size;

      /* seek to a random position in the file - this can be EOF */
      fseek(f, randpos, SEEK_SET);

      fgets(l, n, f);
      /* get line - this can be empty (if we are close to EOF) */
   } while (fgets(l, n, f)==NULL);

   /* close the file */
   fclose(f);
}

void setgamevariable(void)
{
   static char l[MAXLINELEN];
   size_t wordlen;

   if (testword)
   {
	testword=FALSE;		/* generate a random word in the next match */
	tries=stralphalen(word)+errors;
	suggestion="";
   }
   else
   do
   {
      /* for testing purpose replace this line with
		gets(l); */
      getrandline(l, MAXLINELEN);

      word=chomp(my_strtok(l, DELIMITER));
      wordlen=stralphalen(word);
      suggestion=chomp(my_strtok(NULL, DELIMITER));
      tries=atoi(my_strtok(NULL, DELIMITER));
      if (!tries)
	    tries=wordlen+errors;

      /* testing code:
      printf("word=%s\nsuggestion=%s\ntries=%d\n", word, suggestion, tries); */

   } while (chosenlen && wordlen!=chosenlen);

   triesleft=tries;
   triedchars[0]='\0';
}

#define yaddstr(y, add, str) \
        mvaddstr(y, (COLS-strlen(str)-add)/2, (str))

#define reversecurrentline()	chgat(-1, A_REVERSE, 0, NULL)

#define CTRL(x)		((x) & 0x1f)

/* returns TRUE if the player has guessed
   all the word, FALSE otherwise */
bool printunderlinedword(int guesschar)
{
   int i; char c; chtype attr;
   bool guessedall=TRUE;

   move(4, COLS/2-strlen(word));
   for (i=0; (c=word[i]); i++)
   {
      attr=A_NORMAL;
      if (c==guesschar)
	 attr=A_BOLD;
      else if (isalpha(c) && !strcasechr(triedchars, c))
      {
	 c=HIDDENCHAR;
	 guessedall=FALSE;
      }
      addch(c|attr);
      addch(' ');
   }
   return guessedall;
}

/* reveals the right word */
void printword(void)
{
   int i; chtype c;

   move(4, COLS/2-strlen(word));
   for (i=0; (c=word[i]); i++)
   {
      if (!strcasechr(triedchars, c))
         c|=A_BOLD;
      addch(c);
      addch(' ');
   }
}

void printsuggestion(void)
{
   if (!quite && *suggestion)
      mvprintw(7, 0, "suggestion: %s", suggestion);
}

void printtried(void)
{
   char msg[100]; int msglen;

   msglen = sprintf(msg, "tries left: %2d/%-2d", triesleft, tries);
   mvaddstr(9, COLS-msglen, msg);
   mvprintw(9, 0, "tried letters: %s", triedchars);
}

/* prints the title and the help bar */
void printbars(void)
{
   move(0,0);
   reversecurrentline();
   yaddstr(0, 0, " H A N G M A N  game ");
   move(11,0);
   reversecurrentline();
   attron(A_REVERSE);
   yaddstr(11, 0, "^X exit   ^N new game   ^A abort   ^L refresh");
   attrset(A_NORMAL);
}

/* character input routine */
int letter=' ';

int get1str(char *str)
{
   int c;
   short y,x;

   getyx(stdscr, y, x);
   attron(A_BOLD);

   do
   {
        c=tolower(getch());
	if (strchr(triedchars, c))
	   continue;
        else if (islower(c))
        {
           if (letter!=' ')
           {
                move(y,x);
                addch(' ');
                move(y,x);
           }
           letter=c;
           addch(c);
        }
        else if ((c==KEY_BACKSPACE || c==KEY_DC || c=='\b') && letter!=' ')
        {
           move(y,x);
           addch(' ');
           move(y,x);
        }
        else if (c==CTRL('X') || c==CTRL('N') || c==CTRL('A'))
	   break;
	else if (c==CTRL('L'))
            wrefresh(curscr);
   } while (letter==' ' || (c!='\n' && c!=KEY_ENTER));
   attroff(A_BOLD);

   str[0]=letter; str[1]='\0'; 
   letter=' ';
   return c;
}

/* returning true will make a new game to begin,
   false will tell main to exit */
char guesschar[2];

bool playgame(void)
{
   bool win;
   int c;

   attrset(A_NORMAL);
   curs_set(1);		/* normal cursor */
   erase();
   printbars();
   printsuggestion();
   yaddstr(2, 0, lastmsg="welcome to hangman");
   *guesschar='\0';

   for (;;)
   {
      win=printunderlinedword(*guesschar);
      printtried();

      if (!triesleft || win)
      {
	if (win)
	{
	   attron(A_BOLD);
	   yaddstr(2, 0, lastmsg="Good work! You win!");
	}
	else
	{
	   yaddstr(2, 0, lastmsg="Sorry, you've lost. The word was:");
	   printword();
	}

	curs_set(0);	/* hide cursor */
	for (;;)
	{
	   c = getch();
	   switch(c)
	   {
	      case CTRL('X'):
		 return FALSE;
	      case CTRL('N'):
		 return TRUE;
	      case CTRL('L'):
		 wrefresh(curscr);
	   }
	}
      }

      c=get1str(guesschar);
      switch(c)
      {
	 case CTRL('X'):
	    return FALSE;
	 case CTRL('N'):
	    return TRUE;
	 case CTRL('A'):
	    triesleft=0;
	    continue;
      }

      if (strcasechr(word, *guesschar))
	 yaddstr(2, 0, lastmsg="   correct guess!   ");
      else
	 yaddstr(2, 0, lastmsg="sorry, wrong guess.");

      triesleft--;
      strcat(triedchars, guesschar);
   }
}

/* SIGWINCH signal handler */
void resize(int sig)
{
   attr_t attrs;

   endwin();
   refresh();

   /* redraw */
   attr_get(&attrs, (short *)0, NULL);
   attrset(A_NORMAL);
   erase();
   printbars();
   yaddstr(2, 0, lastmsg);
   printsuggestion();
   if (triesleft)
      printunderlinedword(*guesschar);
   else
   {
      printword();
      curs_set(0);
   }
   printtried();
   if (letter!=' ') { attrset(A_BOLD); addch(letter); }
   attr_set(attrs, 0, NULL);
   refresh();
}

void usage(void)
{
   printf("Usage: hangman [OPTION]... [FILE]\n"
	"Play the hangman game with words from FILE.\n\n"

	"  -e, --errors=N       set the number of tries\n"
	"                       to word length+N (default N=1)\n"
	"  -l, --length=N       play with a N char word (N >= 2)\n"
	"  -q, --quite          never display suggestions\n"
	"  -w, --word=WORD      play with this word\n"
	"  -h, --help           display this help and exit\n"
	"  -v, --version        output version information and exit\n\n"
	"Data files will be searched in this order:\n"
	"   [FILE]\t$HOME/" HOMEDATAFILENAME "\t" DATAFILENAME
	"\n\nReport bugs to <antonio.b@infinito.it>.\n");
}

char *prog;     /* contains the name of the program */

int main(int argc, char *argv[])
{
   struct option long_opt[] =
   {
   /*  name             has_arg         	flag    val  */
      {"help",          no_argument,    	NULL,   'h'},
      {"version",       no_argument,    	NULL,   'v'},
      {"quite",		no_argument,		NULL,	'q'},
      {"length",	required_argument,	NULL,	'l'},
      {"errors",	required_argument,	NULL,	'e'},
      {"word",		required_argument,	NULL,	'w'},
      {0, 0, 0, 0}
   };
   int opt;
   int code=EXIT_SUCCESS;
   bool rungame=TRUE;
   prog=argv[0];


   /* command line argument handling */
   while ((opt=getopt_long(argc, argv, "hvql:e:w:", long_opt, NULL)) != -1)
   {
	switch (opt)
	{
	   case 'h':
		usage();
		rungame=FALSE;
		break;
	   case 'v':
		puts(VERSION);
		rungame=FALSE;
		break;
	   case 'q':
		quite=TRUE;
		break;
	   case 'l':
		chosenlen=atoi(optarg);
		if (chosenlen<2)
		{
		   fprintf(stderr, "Please follow -l with a number >=2.\n");
		   code=EXIT_FAILURE;
		}
		break;
	   case 'e':
		errors=atoi(optarg);
		break;
	   case 'w':
		word=optarg;
		testword=TRUE;
		break;
	   case '?':
		fprintf(stderr, "Try `hangman -h' for more information.\n");
		code=EXIT_FAILURE;
	}
   }

   /* process non-options */
   datafile=argv[optind];

   if (code==EXIT_SUCCESS && rungame)
   {
      /* set xterm window resize handler */
      signal(SIGWINCH, resize);
      /* init randomizer */
      srand((unsigned)time(NULL)); 
      /* fire up the game */
      initscr();
      cbreak();
      noecho();
      keypad(stdscr, TRUE);
      do
      {
	 setgamevariable();
      } while (playgame());
      endwin();
   }

   exit(code);
}

