#include <curses.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

/* -----------------------Configure here----------------------- */
/* name of the user data file - this is searched first */
#define HOMEDATAFILENAME        ".hangman.words"
/* name of the default data file */
#define DATAFILENAME    "hangman.words"
/* max length of line of DATAFILENAME */
#define MAXLINELEN	200
/* field delimiter in file DATAFILENAME */
#define DELIMITER	"|"
/* char used to hidden a not yet guessed char */
#define HIDDENCHAR	'_'
/* ------------------------------------------------------------ */

/* game variables */
char *word, *suggestion, triedchars[MAXLINELEN];
int tries, triesleft;
/******************/

/* dies with the name of the program and
   an error message. return the error
   number */
void die(char *what)
{
   extern char* prog;
   fprintf(stderr, "%s: %s\n", prog, what);
   perror(prog);
   exit(errno);
}

/* standard hack to return the file size;
   don't changes the position
   indicator; another solution is to 
   use the unix call stat(2) */
long filesize(FILE *f)
{
   long size, curpos=ftell(f);
   fseek(f, 0, SEEK_END);
   size=ftell(f);
   fseek(f, curpos, SEEK_SET);
   return size;
}

/* removes a newline at the end of a string
   cannot be used for constant strings */
char *chomp(char *s)
{
   if (s && *s)
   {
      size_t last=strlen(s)-1;
      if (s[last]=='\n')
         s[last]='\0';
   }
   return s;
}

/* similar to strchr(3),
   but compares are case-insensitive */
const char *strcasechr(const char *s, int c)
{
   if (s)
      for (c=tolower(c); (tolower(*s) != c); s++)
	if (!*s) return NULL;
   return s;
}

char *my_strtok(char *s, const char *delim)
{
   static char *p;

   for (!s?(s=p):(p=s); !strchr(delim, *p); p++);
   if (*p) *p++='\0';

   return s;
}

void getrandline(char *l, int n)
{
   long size, randpos;
   FILE *f;
   char fname[200];

   /* first try to open $HOME/HOMEDATAFILENAME */
   strcpy(fname, (char *)getenv("HOME"));
   strcat(fname, "/");
   strcat(fname, HOMEDATAFILENAME);
   f=fopen(fname, "r");

   /* if this fails, try to open DATAFILENAME */
   if (f == NULL)
   {
      f=fopen(DATAFILENAME, "r");
      if (f == NULL) die("trying to open " DATAFILENAME);
   }

   size=filesize(f);

   /* init randomizer */
   srand((unsigned)time(NULL));

   do
   {
      /* get a random value in the range [0..size-1] */
      randpos=rand()%size;

      /* seek to a random position in the file - this can be EOF */
      fseek(f, randpos, SEEK_SET);

      fgets(l, n, f);
      /* get line - this can be empty (if we are close to EOF) */
   } while (fgets(l, n, f)==NULL);

   /* close the file */
   fclose(f);
}

void setgamevariable(void)
{
   static char l[MAXLINELEN];

   /* for testing purpose replace this line with
		gets(l); */
   getrandline(l, MAXLINELEN);

   word=chomp(my_strtok(l, DELIMITER));
   suggestion=chomp(my_strtok(NULL, DELIMITER));
   tries=atoi(my_strtok(NULL, DELIMITER));
   if (!tries)
	tries=strlen(word)+1;
   triesleft=tries;

   /* testing code:
   printf("word=%s\nsuggestion=%s\ntries=%d\n", word, suggestion, tries); */
}

#define yaddstr(y, add, str) \
        mvaddstr(y, (COLS-strlen(str)-add)/2, (str))

#define reversecurrentline()	chgat(-1, A_REVERSE, 0, NULL)

void printunderlinedword(void)
{
   int i; char c;

   move(4, COLS/2-strlen(word));
   for (i=0; c=word[i]; i++)
   {
      if (!strcasechr(triedchars, c))
	 c=HIDDENCHAR;
      printw("%c ", c);
   }
}

void printsuggestion(void)
{
   if (*suggestion)
      mvprintw(7, 0, "suggestion: %s", suggestion);
}

void printtried(void)
{
   char msg[100]; int msglen;

   msglen = sprintf(msg, "tries left: %2d/%-2d", triesleft, tries);
   mvaddstr(9, COLS-msglen, msg);
   mvprintw(9, 0, "tried letters: %s", triedchars);
}

/* prints the title and the help bar */
void printbars(void)
{
   move(0,0);
   reversecurrentline();
   yaddstr(0, 0, " H A N G M A N  game ");
   move(11,0);
   reversecurrentline();
   attron(A_REVERSE);
   yaddstr(11, 0, "^X exit   ^N new game   ^A abort   ^L refresh");
   attrset(A_NORMAL);
}

void playgame(void)
{
   initscr();
   printbars();
   printunderlinedword();
   printsuggestion();
   printtried();
   getch();
   endwin();
}

char *prog;     /* contains the name of the program */

int main(int argc, char *argv[])
{
   prog=argv[0];

   setgamevariable();
   playgame();

   exit(0);
}

