#include <curses.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

/* -----------------------Configure here----------------------- */
/* name of the user data file - this is searched first */
#define HOMEDATAFILENAME        ".hangman.words"
/* name of the default data file */
#define DATAFILENAME    "hangman.words"
/* max length of line of DATAFILENAME */
#define MAXLINELEN	200
/* field delimiter in file DATAFILENAME */
#define DELIMITER	"|"
/* char used to hidden a not yet guessed char */
#define HIDDENCHAR	'_'
/* ------------------------------------------------------------ */

/* game variables */
char *word, *suggestion, triedchars[MAXLINELEN];
int tries, triesleft;
/******************/

/* dies with the name of the program and
   an error message. return the error
   number */
void die(char *what)
{
   extern char* prog;
   fprintf(stderr, "%s: %s\n", prog, what);
   perror(prog);
   exit(errno);
}

/* standard hack to return the file size;
   don't changes the position
   indicator; another solution is to 
   use the unix call stat(2) */
long filesize(FILE *f)
{
   long size, curpos=ftell(f);
   fseek(f, 0, SEEK_END);
   size=ftell(f);
   fseek(f, curpos, SEEK_SET);
   return size;
}

/* removes a newline at the end of a string
   cannot be used for constant strings */
char *chomp(char *s)
{
   if (s && *s)
   {
      size_t last=strlen(s)-1;
      if (s[last]=='\n')
         s[last]='\0';
   }
   return s;
}

/* similar to strchr(3),
   but compares are case-insensitive */
const char *strcasechr(const char *s, int c)
{
   if (s)
      for (c=tolower(c); (tolower(*s) != c); s++)
	if (!*s) return NULL;
   return s;
}

char *my_strtok(char *s, const char *delim)
{
   static char *p;

   for (!s?(s=p):(p=s); !strchr(delim, *p); p++);
   if (*p) *p++='\0';

   return s;
}

void getrandline(char *l, int n)
{
   long size, randpos;
   FILE *f;
   char fname[200];

   /* first try to open $HOME/HOMEDATAFILENAME */
   strcpy(fname, (char *)getenv("HOME"));
   strcat(fname, "/");
   strcat(fname, HOMEDATAFILENAME);
   f=fopen(fname, "r");

   /* if this fails, try to open DATAFILENAME */
   if (f == NULL)
   {
      f=fopen(DATAFILENAME, "r");
      if (f == NULL) die("trying to open " DATAFILENAME);
   }

   size=filesize(f);

   /* init randomizer */
   srand((unsigned)time(NULL));

   do
   {
      /* get a random value in the range [0..size-1] */
      randpos=rand()%size;

      /* seek to a random position in the file - this can be EOF */
      fseek(f, randpos, SEEK_SET);

      fgets(l, n, f);
      /* get line - this can be empty (if we are close to EOF) */
   } while (fgets(l, n, f)==NULL);

   /* close the file */
   fclose(f);
}

void setgamevariable(void)
{
   static char l[MAXLINELEN];

   /* for testing purpose replace this line with
		gets(l); */
   getrandline(l, MAXLINELEN);

   word=chomp(my_strtok(l, DELIMITER));
   suggestion=chomp(my_strtok(NULL, DELIMITER));
   tries=atoi(my_strtok(NULL, DELIMITER));
   if (!tries)
	tries=strlen(word)+1;
   triesleft=tries;
   triedchars[0]='\0';

   /* testing code:
   printf("word=%s\nsuggestion=%s\ntries=%d\n", word, suggestion, tries); */
}

#define yaddstr(y, add, str) \
        mvaddstr(y, (COLS-strlen(str)-add)/2, (str))

#define reversecurrentline()	chgat(-1, A_REVERSE, 0, NULL)

#define CTRL(x)		((x) & 0x1f)

/* returns TRUE if the player has guessed
   all the word, FALSE otherwise */
bool printunderlinedword(int guesschar)
{
   int i; char c; chtype attr;
   bool guessedall=TRUE;

   move(4, COLS/2-strlen(word));
   for (i=0; c=word[i]; i++)
   {
      attr=A_NORMAL;
      if (c==guesschar)
	 attr=A_BOLD;
      else if (!strcasechr(triedchars, c))
      {
	 c=HIDDENCHAR;
	 guessedall=FALSE;
      }
      addch(c|attr);
      addch(' ');
   }
   return guessedall;
}

/* reveals the right word */
void printword(void)
{
   int i; chtype c;

   move(4, COLS/2-strlen(word));
   for (i=0; c=word[i]; i++)
   {
      if (!strcasechr(triedchars, c))
         c|=A_BOLD;
      addch(c);
      addch(' ');
   }
}

void printsuggestion(void)
{
   if (*suggestion)
      mvprintw(7, 0, "suggestion: %s", suggestion);
}

void printtried(void)
{
   char msg[100]; int msglen;

   msglen = sprintf(msg, "tries left: %2d/%-2d", triesleft, tries);
   mvaddstr(9, COLS-msglen, msg);
   mvprintw(9, 0, "tried letters: %s", triedchars);
}

/* prints the title and the help bar */
void printbars(void)
{
   move(0,0);
   reversecurrentline();
   yaddstr(0, 0, " H A N G M A N  game ");
   move(11,0);
   reversecurrentline();
   attron(A_REVERSE);
   yaddstr(11, 0, "^X exit   ^N new game   ^A abort   ^L refresh");
   attrset(A_NORMAL);
}

/* character input routine */
int get1str(char *str)
{
   int c, letter=' ';
   short y,x;

   getyx(stdscr, y, x);
   attron(A_BOLD);

   do
   {
        c=tolower(getch());
	if (strchr(triedchars, c))
	   continue;
        else if (islower(c))
        {
           if (letter!=' ')
           {
                move(y,x);
                addch(' ');
                move(y,x);
           }
           letter=c;
           addch(c);
        }
        else if ((c==KEY_BACKSPACE || c==KEY_DC || c=='\b') && letter!=' ')
        {
           move(y,x);
           addch(' ');
           move(y,x);
        }
        else if (c==CTRL('X') || c==CTRL('N') || c==CTRL('A'))
	   break;
	else if (c==CTRL('L'))
            wrefresh(curscr);
   } while (letter==' ' || c!='\n' && c!=KEY_ENTER);
   attroff(A_BOLD);

   str[0]=letter; str[1]='\0'; 
   return c;
}

/* returning true will make a new game to begin,
   false will tell main to exit */
bool playgame(void)
{
   char guesschar[2];
   bool win;
   int c;

   attrset(A_NORMAL);
   curs_set(1);		/* normal cursor */
   erase();
   printbars();
   printsuggestion();
   yaddstr(2, 0, "welcome to hangman");
   *guesschar='\0';

   for (;;)
   {
      win=printunderlinedword(*guesschar);
      printtried();

      if (!triesleft || win)
      {
	if (win)
	{
	   attron(A_BOLD);
	   yaddstr(2, 0, "Good work! You win!");
	}
	else
	{
	   yaddstr(2, 0, "Sorry, you've lost. The word was:");
	   printword();
	}

	curs_set(0);	/* hide cursor */
	for (;;)
	{
	   c = getch();
	   switch(c)
	   {
	      case CTRL('X'):
		 return FALSE;
	      case CTRL('N'):
		 return TRUE;
	      case CTRL('L'):
		 wrefresh(curscr);
	   }
	}
      }

      c=get1str(guesschar);
      switch(c)
      {
	 case CTRL('X'):
	    return FALSE;
	 case CTRL('N'):
	    return TRUE;
	 case CTRL('A'):
	    triesleft=0;
	    continue;
      }

      if (strcasechr(word, *guesschar))
	 yaddstr(2, 0, "   correct guess!   ");
      else
	 yaddstr(2, 0, "sorry, wrong guess.");

      triesleft--;
      strcat(triedchars, guesschar);
   }
}

char *prog;     /* contains the name of the program */

int main(int argc, char *argv[])
{
   prog=argv[0];

   initscr();
   cbreak();
   noecho();
   keypad(stdscr, TRUE);
   do
   {
      setgamevariable();
   } while (playgame());
   endwin();

   exit(0);
}

